/*
 *    Columbia University     
 *    Computer Science Department
 *
 *    Andrew Lih (lih@cs.columbia.edu)
 *    Erez Zadok <ezk@cs.columbia.edu>
 *    May 1993
 *
 *    pvm3_ad.c: implements the GNU make agent daemon for PVM tasks
 *
 *    This daemon is called via pvm_spawn() without any arguments.
 *
 *    After starting, this daemon expects a message buffer to be passed
 *    to it, containing the following information:
 *
 *       char*    working directory
 *       char**   argv
 *       char**   envp
 *
 *    The daemon then forks and execve's the command in argv[0] with the
 *    rest of the argv vector and envp.  stdout and stderr are redirected
 *    to files in /tmp.  After the child finishes, the following are sent
 *    back to the pvm invoker:
 *
 *       int      tid
 *       int      status
 *       char*    stdout
 *       char*    stderr
 *
 *    Now complete with all its actions, this process then dies.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <signal.h>

#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "pvm3.h"
#include "pvm3_ad.h"
#include "pgmake.h"

static char rcsid[] = "$Id: pvm3_ad.c,v 1.6 1993/05/13 16:21:05 lih Exp lih $";

static int debug = 0;		/* no debugging by default */

main()
{
  char *resolve_path();

  int mytid;			/* This task id */
  int pid;

  char	*wd;			/* Working directory */
  int    argc;			/* ARG stuff */
  char **argv;
  int    envc;			/* ENV stuff */
  char **envp;

  int ourpid;
  int status = 0;
  int i, master;
  char buf [BUFSIZ];
  char **cp;

  int   outfd, errfd;
  int   outpipe[2], errpipe[2];

  char  outfile  [MAX_TEMP_PATH_LEN];
  char  errfile  [MAX_TEMP_PATH_LEN];
  char  filename [MAX_TEMP_PATH_LEN];
    
  FILE *ofile;

  /* Get current PID
   */
  ourpid = getpid();
  master = pvm_parent();

  sprintf (filename, "/tmp/pvm3_ad.%d", getpid());
  ofile = fopen (filename, "w");

  /* enroll with pvmd */
  mytid = pvm_mytid();

  if (debug) {
    fprintf(ofile, "*** \n");
    fprintf(ofile, "maked: tid = %x\n",mytid);
  }

  /* Receive data from master */
  pvm_recv( -1, PGMAKE_JOB_BEGIN );

  if (debug) {
    fprintf(ofile, "maked: pvm_recv succeeded\n");
  }

  /* Unpack debugging flag from GNU make.  Of course every debugging-
     related option would not print until the flag is determined now... */
  pvm_upkint(&debug, 1, 1);

  /* Unpack working directory */
  pvm_upkstr(buf);
  wd = (char*) malloc (sizeof(char) * (strlen(buf)+1));
  if (wd == NULL && debug) {
    fprintf(stderr, "no more memory for pvmd3_ad\n");
    exit(2);
  }
  strcpy(wd, buf);

  /* Unpack arg vector */
  pvm_upkint(&argc, 1, 1);
  argv = (char **) malloc (sizeof(char*) * (argc+1));
  if (argv == NULL && debug) {
    fprintf(stderr, "no more memory for pvmd3_ad\n");
    exit(2);
  }
  for (i=0; i<argc; i++) {
    pvm_upkstr(buf);
    argv[i] = (char*) malloc (sizeof(char) * (strlen(buf)+1));
    if (argv[i] == NULL && debug) {
      fprintf(stderr, "no more memory for pvmd3_ad\n");
      exit(2);
    }
    strcpy(argv[i], buf);
  } 
  argv[i] = NULL;

  /* Unpack environment */
  pvm_upkint(&envc, 1, 1);
  envp = (char **) malloc (sizeof(char*) * (envc+1));
  if (envp == NULL && debug) {
    fprintf(stderr, "no more memory for pvmd3_ad\n");
    exit(2);
  }

  for (i=0; i<envc; i++) {
    pvm_upkstr(buf);
    envp[i] = (char*) malloc (sizeof(char) * (strlen(buf)+1));
    if (envp[i] == NULL && debug) {
      fprintf(stderr, "no more memory for pvmd3_ad\n");
      exit(2);
    }
    strcpy(envp[i], buf);
  }      
  envp[i] = NULL;

  /* Print out what I am doing
   */
  if (debug) {
    /* print working directory */
    fprintf(ofile, "maked: wd = %s\n", wd);
    /* print argv */
    fprintf(ofile, "maked: argv = ");
    for (cp = argv; cp && *cp != NULL; cp++)
      fprintf (ofile, "%s ", *cp);
    fprintf (ofile, "\n\n");
    /* print envp */
    fprintf(ofile, "maked: envp = ");
    for (cp = envp; cp && *cp != NULL; cp++)
      fprintf (ofile, "%s ", *cp);
    fprintf (ofile, "\n\n");
  }

  /*** fork and execve section
   ***    do the real work
   ***/
#ifdef FUZCOMMENT
  sprintf (outfile, STDOUT_FILENAME, ourpid);
  sprintf (errfile, STDERR_FILENAME, ourpid);

  outfd = open (outfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  errfd = open (errfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    
  if (dup2 (outfd, STDOUT_FILENO) < 0 && debug)
    perror("dup2");

  if (dup2 (errfd, STDERR_FILENO) < 0 && debug)
    perror("dup2");
#endif

  if (pipe(outpipe)<0 || pipe(errpipe)<0)
    perror("pipe");

  fprintf(ofile,"pipe: returned outpipe[2] = {%d, %d}\n",
	  outpipe[0], outpipe[1]);
  fprintf(ofile,"pipe: returned errpipe[2] = {%d, %d}\n",
	  errpipe[0], errpipe[1]);

  if ( (pid = fork()) < 0) {
    if (debug)
      perror("fork");
    exit (1);
  }

  if (pid > 0) {

    int    retval = 0;
    fd_set rset;
    struct rusage  usage;
    struct timeval tv;

    int n;
    char bigbuf[LARGE_BUFSIZ];

    /*   Parent code: loop non-blockingly (sp!) to check whether either
     *   the child has exited or if the parent PVM process sent a kill
     *   message.
     */

    close (outpipe[1]);
    close (errpipe[1]);

    tv.tv_sec  = 0;		/* Set blocking time */
    tv.tv_usec = 0;

    FD_ZERO(&rset);
    FD_SET(outpipe[0], &rset);
    FD_SET(errpipe[0], &rset);

    while (!retval) {

      if (debug)
	fprintf(ofile,"while: looping\n", pid);

      /*****
       *****  Select on the file descriptors for stderr/stdout
       *****    coming from the child process.
       *****/
      if (select (10, &rset, NULL, NULL, &tv)) {
	if (debug)
	  fprintf(ofile,"select: inside select\n", pid);
	if (FD_ISSET(outpipe[0], &rset)) {
	  pvm_initsend( PvmDataDefault );
	  n = read(outpipe[0], bigbuf, LARGE_BUFSIZ);
	  bigbuf[n] = '\0';
	  pvm_pkstr(bigbuf);
	  pvm_send(master, PGMAKE_STDOUT);
	}
	if (FD_ISSET(errpipe[0], &rset)) {
	  pvm_initsend( PvmDataDefault );
	  n = read(errpipe[0], bigbuf, LARGE_BUFSIZ);
	  bigbuf[n] = '\0';
	  pvm_pkstr(bigbuf);
	  pvm_send(master, PGMAKE_STDERR);
	}
      }

      /* Process kill request from anyone */
      if (pvm_nrecv(-1 , PGMAKE_KILL)) {
	int signal;
	pvm_upkint (&signal, 1, 1);	/* Unpack the id     */
	if (debug)
	  fprintf(ofile,">>> KILLING child %d, due to kill message\n", pid);
	kill (pid, signal);
      }

      retval = wait3 (&status, WNOHANG, &usage );
    }

    if (retval < 0) {
      if (debug)
	perror("pvm3/wait3");
    } else if (!retval) {
      if (debug)
	fprintf(ofile,">>> KILLED child %d, due to kill message\n", pid);
    } else {
      if (debug)
	fprintf(ofile,">>> WAIT3 returned %d, status %d\n", retval, status);
    }
   
  } else {
      
    /*   Child code:
     *   First chdir() to the right directory
     *   Do an execve with the argv and envp
     */
    close (outpipe[0]);
    close (errpipe[0]);

    if (outpipe[1]!=STDOUT_FILENO)
      if (dup2 (outpipe[1], STDOUT_FILENO) < 0 && debug)
	perror("dup2");

    if (errpipe[1]!=STDERR_FILENO)
      if (dup2 (errpipe[1], STDERR_FILENO) < 0 && debug)
	perror("dup2");

    close (outpipe[1]);
    close (errpipe[1]);

    if (chdir(wd) < 0) {
      if (debug)
	fprintf (ofile,"maked: child couldn't chdir to %s\n",wd); 
      status = -1;
    }
    if (!status) {
      char *p;
      if ((p = resolve_path(argv[0], envp)) == NULL)
	execve(argv[0], argv, envp);
      else {
	execve(p, argv, envp);
      }
    }
  }

  /* Send result to master */
  pvm_initsend( PvmDataDefault );
  pvm_pkint( &mytid,  1, 1 );
  pvm_pkint( &status, 1, 1 );

#ifdef FUZCOMMENT
  /* Possible room for improvement: dynamic sizing buffer
   */
  {
    int n;
    char bigbuf[LARGE_BUFSIZ];

    lseek (outfd, 0, SEEK_SET);
    n = read(outfd, bigbuf, LARGE_BUFSIZ);
    bigbuf[n] = '\0';
    pvm_pkstr(bigbuf);

    lseek (errfd, 0, SEEK_SET);
    n = read(errfd, bigbuf, LARGE_BUFSIZ);
    bigbuf[n] = '\0';
    pvm_pkstr(bigbuf);
  }
#endif

  if (master == PvmNoParent) {
    if (debug)
      fprintf(ofile,">>> not forked by pvm_spawn\n");
  }
  pvm_send( master, PGMAKE_JOB_END );

  fclose(ofile);
  close (outfd);
  close (errfd);

  /* leave the debugging/output files around if debugging */
  if (!debug) {
    unlink(filename);
    unlink(outfile);
    unlink(errfile);
  }

  /* Program finished. Exit PVM before stopping */
  pvm_exit();
  exit(0);
}


/*****************************************************************************
 * The following functions were taken out of the sources for an older        *
 * project.								     *
 *****************************************************************************/


/*
 * Search the path for the p program.  NULL if not found.  Full path if
 * found.
 */
char *resolve_path(p, envp)
     char *p, **envp;
{ static char *vn, s[MAX_TEMP_PATH_LEN], *tmp, **s2;
  char *dot;
				/* If path contains '/' return it */
  if (strstr(p, "/"))
    return p;
				/* get PATH from envp that was passed */
  s[0] = '\0';
  s2 = NULL;
  for (s2 = envp; *s2 != NULL; ++s2)
    if (strncmp(*s2, "PATH=", 5) == 0) {
      strcpy(s, &(*s2)[5]);
      break;
    }
  if (*s2 == NULL) {
    if (debug)
      fprintf(stderr, "%s: not found\n", s);
    /* must do something more intelligent */
  }

  tmp = (char *) calloc(MAX_TEMP_PATH_LEN, sizeof(char));
  if (!tmp) {
    if (debug)
      fprintf(stderr, "no more memory for resolve_path\n");
    exit(-1);
  }
				/* scan path for ':' delimiters */
  dot = strstr(s, "::");	/* strtok skips these and we need it */
  vn = strtok(s, ":");
  if (dot  &&  dot < vn) {
    strcpy(tmp, "./");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    dot = NULL;
  }
  while (vn != NULL) {
    if (dot  &&  dot < vn) {
      strcpy(tmp, "./");
      strcat(tmp, p);
      if (!access(tmp, X_OK)) {
	return tmp;
      }
      dot = NULL;
    }
    strcpy(tmp, vn);
    strcat(tmp, "/");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    vn = strtok(NULL, ":");
  }
  if (!vn && dot) {	/* last possibility is that :: is last in PATH */
    strcpy(tmp, "./");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    dot = NULL;
  }
  free(tmp);
  tmp = (char *) NULL;
  return (char *) NULL;
}
