/*
 *    Columbia University     
 *    Computer Science Department
 *
 *    Andrew Lih (lih@cs.columbia.edu)
 *    May 1993
 *
 *    maked.c: implements the make daemon for PVM tasks
 *
 *    This daemon is called via pvm_spawn() without any arguments.
 *
 *    After starting, this daemon expects a message buffer to be passed
 *    to it, containing the following information:
 *
 *       char*    working directory
 *       char**   argv
 *       char**   envp
 *
 *    The daemon then forks and execve's the command in argv[0] with the
 *    rest of the argv vector and envp.  stdout and stderr are redirected
 *    to files in /tmp.  After the child finishes, the following are sent
 *    back to the pvm invoker:
 *
 *       int      tid
 *       int      status
 *       char*    stdout
 *       char*    stderr
 *
 *    Now complete with all its actions, this process then dies.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "pvm3.h"

#include "maked.h"

static char rcsid[]=
"$Id: maked.c,v 1.3 1993/05/01 02:39:37 lih Exp lih $";

main()
{
    int mytid;			/* This task id */
    int pid;

    char  *wd;			/* Working directory */
    int    argc;		/* ARG stuff */
    char **argv;
    int    envc;		/* ENV stuff */
    char **envp;

    int ourpid;
    int i, master, msgtype;
    int status=0;
    char buf [BUFSIZ];

    int   outfd, errfd;
    char  outfile  [MAX_TEMP_PATH_LEN];
    char  errfile  [MAX_TEMP_PATH_LEN];
    char  filename [MAX_TEMP_PATH_LEN];
    
    FILE *ofile;

    /* Get current PID
     */
    ourpid = getpid();

    sprintf (filename,"/tmp/maked.%d",getpid());
    ofile = fopen (filename,"w");

    /* enroll in pvm */
    mytid = pvm_mytid();

    fprintf(ofile,"*** \n");
    fprintf(ofile,"maked: tid = %x\n",mytid);

    /* Receive data from master */
    msgtype = 0;
    pvm_recv( -1, msgtype );

    fprintf(ofile,"maked: pvm_recv succeeded\n");

    /* Unpack working directory */
    pvm_upkstr(buf);
    wd = (char*) malloc (sizeof(char) * (strlen(buf)+1));
    strcpy(wd, buf);

    /* Unpack arg vector */
    pvm_upkint(&argc, 1, 1);
    argv = (char **) malloc (sizeof(char*) * (argc+1));
    for (i=0; i<argc; i++) {
      pvm_upkstr(buf);
      argv[i] = (char*) malloc (sizeof(char) * (strlen(buf)+1));
      strcpy(argv[i], buf);
    } 
    argv[i]="";

    /* Unpack environment */
    pvm_upkint(&envc, 1, 1);
    envp = (char **) malloc (sizeof(char*) * (envc+1));
    for (i=0; i<envc; i++) {
      pvm_upkstr(buf);
      envp[i] = (char*) malloc (sizeof(char) * (strlen(buf)+1));
      strcpy(envp[i], buf);
    }      
    envp[i]="";

    /* Print out what I am doing
     */
    fprintf(ofile,"maked: wd = %s\n", wd);

    fprintf(ofile,"maked: argv = ");
    {
      char **cp;
      for (cp=argv; *cp!=NULL; cp++)
	fprintf (ofile,"%s ", *cp);
      fprintf (ofile,"\n\n");
    }

    fprintf(ofile,"maked: envp = ");
    {
      char **cp;
      for (cp=envp; *cp!=NULL; cp++)
	fprintf (ofile,"%s ", *cp);
      fprintf (ofile,"\n\n");
    }

    /*** fork and execve section
     ***    do the real work
     ***/
    sprintf (outfile, STDOUT_FILENAME, ourpid);
    sprintf (errfile, STDERR_FILENAME, ourpid);

    outfd = open (outfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    errfd = open (errfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    
    if (dup2 (outfd, STDOUT_FILENO)<0)
      perror("dup2");

    if (dup2 (errfd, STDERR_FILENO)<0)
      perror("dup2");

    if ( (pid=fork()) <0) {
      perror("fork");
      exit (1);
    }

    if (pid > 0) {

      /*   Parent code:
       *   eventually we should change this to use wait3()
       *   and receive messages from the master via pvm_nrecv()
       *   in case the other side wants to kill jobs.
       */
      wait (&status);

    } else {
      
      /*   Child code:
       *   First chdir() to the right directory
       *   Do an execve with the argv and envp
       */
      if (chdir(wd)) {
	fprintf (ofile,"maked: child couldn't chdir to %s\n",wd); 
	status=-1;
      }
      if (!status)
	execve(argv[0], argv, envp);
    }

    /* Send result to master */
    pvm_initsend( PvmDataDefault );
    pvm_pkint( &mytid,  1, 1 );
    pvm_pkint( &status, 1, 1 );

    /* Possible room for improvement: dynamic sizing buffer
     */
    {
      int n;
      char bigbuf[LARGE_BUFSIZ];

      lseek (outfd, 0, SEEK_SET);
      n = read(outfd, bigbuf, LARGE_BUFSIZ);
      bigbuf[n]='\0';
      pvm_pkstr(bigbuf);

      lseek (errfd, 0, SEEK_SET);
      n = read(errfd, bigbuf, LARGE_BUFSIZ);
      bigbuf[n]='\0';
      pvm_pkstr(bigbuf);
    }

    master = pvm_parent();
    pvm_send( master, 0 );

    fclose(ofile);
    close (outfd);
    close (errfd);

    /*
    unlink(outfile);
    unlink(errfile);
    */

    /* Program finished. Exit PVM before stopping */
    pvm_exit();
}
